type point = { x : int; y : int }

type package_location =
    Carried of int
  | Available of point option

type package = { package_id : int;
		 mutable weight : int option;
		 mutable dest : point option;
		 mutable package_pos : package_location option }
type robot = { robot_id : int;
	       mutable robot_pos : point option }

type tile_type = Normal | Home | Water | Wall
type tile = { loc : point;
	      ty : tile_type;

	      (* dijkstra-related fields *)
	      mutable depth : int;
	      mutable pred : tile option;
	      mutable in_q : bool
	    }

let out = function Some(x) -> x | None -> assert false

let list_of_string s =
  let rec inner sofar i =
    (match i with
	 -1 -> sofar
       | _ -> inner (s.[i] :: sofar) (i - 1))
  in
    inner [] ((String.length s) - 1)

let string_of_point pt =
  "(" ^ (string_of_int pt.x) ^ "," ^ (string_of_int pt.y) ^ ")"

let read_int_last_sep_char = ref ' '
let read_int chan =
  let had_one = ref false in
  let rec inner sofar =
    let c = input_char chan in
      if c >= '0' && c <= '9' then
	(had_one := true;
	 inner (sofar * 10 + ((Char.code c) - (Char.code '0'))))
      else if !had_one then
	(read_int_last_sep_char := c;
	 sofar)
      else
	raise Not_found
  in
    inner 0

let read_map chan =
  let width = read_int chan in
  let height = read_int chan in
    if (width <= 0) then failwith "invalid map width";
    if (height <= 0) then failwith "invalid map height";
    let rec read_lines sofar n =
      match n with
	  0 -> sofar
	| _ -> 
	    let line = input_line chan in
	      if not ((String.length line) = width) then failwith "invalid map data";
	      read_lines (line :: sofar) (n - 1)
    in
      (width,height,List.rev (read_lines [] height))

let load_map (width,height,rows) =
  let load_row _y a_row =
    let _x = ref 0 in
      Array.of_list
	(List.map
	   (fun tile_char ->
	      incr _x;
	      { loc = { x = !_x; y = !_y; };
		ty = (match tile_char with
			  '.' -> Normal
			| '@' -> Home
			| '~' -> Water
			| '#' -> Wall
			| _ -> failwith "invalid map data");
		depth = max_int; (* I use max_int as infinity throughout, trusting no path length will be larger. *)
		pred = None;
		in_q = true;
	      })
	   (list_of_string a_row))
  in
  let _y = ref 0 in
    Array.of_list (List.map (fun a_row -> incr _y; load_row _y a_row) rows)

let print_map map =
  Array.iter
    (fun row ->
       Array.iter (fun a_tile ->
		     (match a_tile.ty with
			  Normal -> print_char '.'
			| Home -> print_char '@'
			| Water -> print_char '~'
			| Wall -> print_char '#')) row;
       print_char '\n')
    map

let lookup_tile map loc =
  Array.get (Array.get map (loc.y - 1)) (loc.x - 1)

(* gets the list of tiles adjacent to loc *)
let adj (width,height,map) loc =
  let f = (fun loc' ->
	     if loc'.x > 0 && loc'.x <= width && loc'.y > 0 && loc'.y <= height then
	       Some(lookup_tile map loc')
	     else
	       None) in
    List.map out
      (List.filter (function Some(x) -> true | _ -> false)
	 [ f { x = pred loc.x; y = loc.y };
	   f { x = succ loc.x; y = loc.y };
	   f { x = loc.x; y = pred loc.y };
	   f { x = loc.x; y = succ loc.y }])

let reachable a_tile =
  a_tile.ty = Normal || a_tile.ty = Home

(* gets the list of reachable tiles adjacent to loc *)
let reachable_adj m loc =
  List.filter reachable (adj m loc)

(* Dijkstra crap *)

let initialize_single_source (width,height,map) source_loc =
  for i = 0 to (height - 1) do
    let row = Array.get map i in
      for j = 0 to (width - 1) do
	let a_tile = Array.get row j in
	  if i + 1 = source_loc.y && j + 1 = source_loc.x then
	    a_tile.depth <- 0
	  else
	    a_tile.depth <- max_int;
	  a_tile.pred <- None;
	  a_tile.in_q <- true
      done
  done

let relax u v =
  if v.depth > succ u.depth then
    (v.depth <- succ u.depth;
     v.pred <- Some(u)
    )

(* searches the map for the tile with the smallest depth.
   this sucks, but it's the best we can do without converting the matrix
   to a fibonacci queue that gets updated during relaxation.
*)
let min_depth (width,height,map) =
  let sofar = ref None in
  let sofar_depth = ref max_int in
    for i = 0 to pred height do
      let row = Array.get map i in
	for j = 0 to pred width do
	  let a_tile = Array.get row j in
	    if a_tile.in_q && a_tile.depth < !sofar_depth then
	      (sofar := Some(a_tile);
	       sofar_depth := a_tile.depth)
	done
    done;
    if !sofar = None then
      raise Not_found
    else
      out !sofar

let dijkstra (width,height,map) source_loc =
  initialize_single_source (width,height,map) source_loc;
  let num_in_q = ref (width*height) in
    try
      while !num_in_q > 0 do
	let u = min_depth (width,height,map) in
	  u.in_q <- false;
	  decr num_in_q;
	  List.iter (fun v -> relax u v) (reachable_adj (width,height,map) u.loc)
      done
    with
	Not_found -> () (* only unreachable tiles remain *)

exception No_path
(* given a pre-dijkstra'd map, returns the shortest path between the source (of the dijkstra run)
   and the given destination tile, as a list of points. *)
let path dst =
  let rec inner cur sofar =
    (if cur.depth = 0 then
       cur.loc :: sofar
     else if cur.pred = None then
       raise No_path
     else
       inner (out cur.pred) (cur.loc :: sofar)) in
    inner dst []

type direction =
    North
  | South
  | East
  | West

type command =
    Move of direction
  | Pick of int list
  | Drop of int list

type bid = int * command

let string_of_bid (b,c) =
  (string_of_int b) ^
  (match c with
       Move(d) ->
	 " Move " ^
	 (match d with
	      North -> "N"
	    | South -> "S"
	    | East -> "E"
	    | West -> "W")
     | Pick(pkgs) ->
	 " Pick" ^
	 List.fold_left (fun rest id ->
			   rest ^ " " ^ (string_of_int id)) "" pkgs
     | Drop(pkgs) ->
	 " Drop" ^
	 List.fold_left (fun rest id ->
			   rest ^ " " ^ (string_of_int id)) "" pkgs)

type state_update =
    Moved of direction
  | Picked of int
  | Dropped of int
  | Appeared of point

(* multi-turn plan *)
type plan =
    Go of point * point list (* go to dest via path *)
  | Attack of int (* attack this robot *)

type state =
    {
      width : int;
      height : int;
      map : tile array array;
      my_id : int;
      mutable my_pos : point;
      mutable my_packages : int list;
      capacity : int;
      mutable money : int;
      packages : (int,package) Hashtbl.t;
      robots : (int,robot) Hashtbl.t;
      mutable my_plan : plan;
      mutable replan : bool;
      mutable unvisited_homebases : point list;
      mutable score : int
    }

let offset_point p d =
  (match d with
       North -> { x = p.x; y = succ p.y }
     | South -> { x = p.x; y = pred p.y }
     | East -> { x = succ p.x; y = p.y }
     | West -> { x = pred p.x; y = p.y })

let update_my_pos s new_pos =
  print_string ("My position: " ^ string_of_point new_pos ^ "\n");
  s.my_pos <- new_pos;
  s.unvisited_homebases <- List.filter (fun pt -> not (pt = s.my_pos)) s.unvisited_homebases

let path_to_package s pkg =
  (match pkg.package_pos with
       Some(Available(Some(pt))) ->
	 let the_tile = lookup_tile s.map pt in
	   path the_tile
     | _ -> raise No_path)

let path_to_package_dest s pkg =
  (match pkg.dest with
       Some(pt) ->
	 let the_tile = lookup_tile s.map pt in
	   path the_tile
     | _ -> raise No_path)

let nearest_dests s packages =
  List.sort
    (fun pkg1 pkg2 ->
       let len1 = List.length (path_to_package_dest s pkg1) in
       let len2 = List.length (path_to_package_dest s pkg2) in
	 if len1 > len2 then 1 else if len1 < len2 then -1 else 0)
    (List.filter
       (fun pkg ->
	  try
	    ignore (path_to_package_dest s pkg);
	    true
	  with
	      Not_found | No_path -> false) packages)

let nearest_pickable_package s =
  let package_list = Hashtbl.fold (fun id pkg sofar -> pkg :: sofar) s.packages [] in
  let nearest =
    List.sort
      (fun pkg1 pkg2 ->
	   let len1 = List.length (path_to_package s pkg1) in
	   let len2 = List.length (path_to_package s pkg2) in
	     if len1 > len2 then 1 else if len1 < len2 then -1 else 0)
      (List.filter
	 (fun pkg ->
	    try
	      ignore (path_to_package s pkg);
	      true
	    with
		Not_found | No_path -> false) package_list) in
    if List.length nearest > 0 then
      match (List.hd nearest).package_pos with
	  Some(Available(Some(pt))) -> pt
	| _ -> assert false
    else
      raise Not_found

let nearest_package_destination s =
  let pkgs = nearest_dests s (List.map (fun id -> Hashtbl.find s.packages id) s.my_packages) in
    if List.length pkgs > 0 then
      out (List.hd pkgs).dest
    else
      raise Not_found

let nearest_homebase s =
  let tiles = List.map (fun pt -> lookup_tile s.map pt) s.unvisited_homebases in
  let nearest = List.sort
		  (fun tile1 tile2 ->
		     let len1 = List.length (path tile1) in
		     let len2 = List.length (path tile2) in
		       if len1 > len2 then 1 else if len1 < len2 then -1 else 0)
		  (List.filter
		     (fun tile ->
			try
			  ignore (path tile);
			  true
			with
			    No_path -> false) tiles) in
    if List.length nearest > 0 then
      (List.hd nearest).loc
    else
      raise Not_found

let rec random_reachable_point s =
  let pt = { x = (Random.int s.width) + 1; y = (Random.int s.height) + 1 } in
    try
      let tile = lookup_tile s.map pt in
	ignore (path tile);
	pt
    with
	Not_found | No_path -> random_reachable_point s

let make_plan s pkgs_here =
  dijkstra (s.width,s.height,s.map) s.my_pos;
  let pt =
    (try
       (try
	  (try
	     nearest_package_destination s
	   with
	       Not_found -> nearest_pickable_package s)
	with
	    Not_found -> nearest_homebase s)
     with
	 Not_found -> random_reachable_point s)
  in
    print_string ("New plan: go to " ^ (string_of_point pt) ^ "\n\n");
    s.my_plan <- Go(pt,path (lookup_tile s.map pt))

let adj_points_offset src dst =
  (match (dst.x - src.x,dst.y - src.y) with
       (0,1) -> North
     | (0,-1) -> South
     | (1,0) -> East
     | (-1,0) -> West
     | _ -> raise (Invalid_argument "adj_points_offset"))

let available_capacity s =
  s.capacity - (List.fold_left (fun sofar pkg_id ->
				  let pkg = Hashtbl.find s.packages pkg_id in
				    sofar + (out pkg.weight)) 0 s.my_packages)

let pickables s (pkgs_here:package list) =
  let avail = available_capacity s in
  let rec inner sofar used nearest =
    (match nearest with
	 hd :: tl ->
	   if used + (out hd.weight) > avail then
	     inner sofar used tl
	   else
	     inner (hd.package_id :: sofar) (used + (out hd.weight)) tl
       | [] -> sofar) in
    inner [] 0 (nearest_dests s pkgs_here)

let deliverables s =
  List.filter (fun id ->
		 let pkg = Hashtbl.find s.packages id in
		   (out pkg.dest) = s.my_pos) s.my_packages

let rec execute_plan s pkgs_here : bid =
  if s.replan then
    (make_plan s pkgs_here;
     s.replan <- false);
  let drops = deliverables s in
    if List.length drops > 0 then
      (1,Drop drops)
    else
      let picks = pickables s pkgs_here in
	if (List.length picks > 0) then
	  (1,Pick picks)
	else
	  (match s.my_plan with
	       Go(pt,path) ->
		 if s.my_pos = pt || List.length path <= 1 then
		   (print_string "Arrived.\n";
		    s.replan <- true;
		    execute_plan s pkgs_here)
		 else
		   (match path with
			fst :: snd :: rest ->
			  if fst = s.my_pos then
			    (s.my_plan <- Go(pt,snd :: rest);
			     (1,Move (adj_points_offset s.my_pos snd)))
			  else
			    (s.replan <- true;
			     execute_plan s pkgs_here)
		      | _ -> assert false)
	     | _ -> assert false)
	  
let update_package_state s pkgs_here =
  List.iter
    (fun a_pkg ->
       try
	 let the_pkg = Hashtbl.find s.packages a_pkg.package_id in
	   the_pkg.weight <- a_pkg.weight;
	   the_pkg.dest <- a_pkg.dest;
	   the_pkg.package_pos <- Some(Available(Some(s.my_pos)))
       with
	   Not_found ->
	     a_pkg.package_pos <- Some(Available(Some(s.my_pos)));
	     Hashtbl.add s.packages a_pkg.package_id a_pkg)
    pkgs_here;
  Hashtbl.iter (fun id a_pkg ->
		  (match a_pkg.package_pos with
		       Some(Available(Some(pt))) when pt = s.my_pos ->
			 (try
			    ignore (List.find (fun pkg -> pkg.package_id = id) pkgs_here)
			  with
			      Not_found -> a_pkg.package_pos <- None)
		     | _ -> ())) s.packages

let update_robot_state s bot msg =
  let me = bot.robot_id = s.my_id in
    (match msg with
	 Moved(d) ->
	   bot.robot_pos <-
	   if not (bot.robot_pos = None) then
	     Some(offset_point (out bot.robot_pos) d)
	   else 
	     None;
	   if me && not (bot.robot_pos = None) then
	     update_my_pos s (out bot.robot_pos)
       | Picked(p) ->
	   (try
	      let pkg = Hashtbl.find s.packages p in
		pkg.package_pos <- Some(Carried(bot.robot_id))
	    with
		Not_found ->
		  let pkg = { package_id = p;
			      weight = None;
			      dest = None;
			      package_pos = Some(Carried(bot.robot_id)) } in
		    Hashtbl.add s.packages p pkg);
	   if me then
	     (s.my_packages <- p :: s.my_packages;
	      s.replan <- true)
     | Dropped(p) ->
	 (try
	    let pkg = Hashtbl.find s.packages p in
	      (match (pkg.dest,bot.robot_pos) with
		   (Some pt1, Some pt2) when pt1 = pt2 ->
		     Hashtbl.remove s.packages p;
		     if me then
		       s.score <- s.score + (out pkg.weight)
		 | _ -> pkg.package_pos <- Some(Available(bot.robot_pos)))
	  with
	      Not_found ->
		let pkg = { package_id = p;
			    weight = None;
			    dest = None;
			    package_pos = Some(Available(bot.robot_pos)) } in
		  Hashtbl.add s.packages p pkg);
	 if me then
	   (s.my_packages <- List.filter (fun a_pkg -> not (a_pkg = p)) s.my_packages;
	    s.replan <- true)
     | Appeared(loc) ->
	 bot.robot_pos <- Some(loc);
 	 if me then
	   update_my_pos s (out bot.robot_pos))
      
let make_bid s ((b,c) as the_bid) chan =
  s.money <- s.money - b;
  print_string ("My money: " ^ (string_of_int s.money) ^ "\n");
  output_string chan (string_of_bid the_bid);
  output_char chan '\n';
  flush chan

let rec nextc chan =
  let c = input_char chan in
    if c = ' ' then
      nextc chan
    else
      c

let read_updates chan =
  let rec inner all_sofar sofar id last_read_int =
    if last_read_int && !read_int_last_sep_char = '\n' then
      List.rev ((id,List.rev sofar) :: all_sofar)
    else
      (match nextc chan with
	   '\n' ->
	     List.rev ((id,List.rev sofar) :: all_sofar)
	 | '#' ->
	     inner ((id,List.rev sofar) :: all_sofar) [] (read_int chan) false
	 | 'N' ->
	     print_string ((string_of_int id) ^ ": Move North\n");
	     inner all_sofar (Moved(North) :: sofar) id false
	 | 'S' ->
	     print_string ((string_of_int id) ^ ": Move South\n");
	     inner all_sofar (Moved(South) :: sofar) id false
	 | 'E' ->
	     print_string ((string_of_int id) ^ ": Move East\n");
	     inner all_sofar (Moved(East) :: sofar) id false
	 | 'W' ->
	     print_string ((string_of_int id) ^ ": Move West\n");
	     inner all_sofar (Moved(West) :: sofar) id false
	 | 'P' ->
	     if not (input_char chan = ' ') then failwith "bad input from server";
	     print_string ((string_of_int id) ^ ": Pick\n");
	     inner all_sofar (Picked(read_int chan) :: sofar) id true
	 | 'D' ->
	     if not (input_char chan = ' ') then failwith "bad input from server";
	     print_string ((string_of_int id) ^ ": Drop\n");
	     inner all_sofar (Dropped(read_int chan) :: sofar) id true
	 | 'X' ->
	     if not (input_char chan = ' ') then failwith "bad input from server";
	     let _x = read_int chan in
	       (if not (nextc chan = 'Y') then failwith "bad input from server";
		if not (input_char chan = ' ') then failwith "bad input from server";
		let _y = read_int chan in
		  (print_string ((string_of_int id) ^ ": Appear " ^ (string_of_int _x) ^ "," ^ (string_of_int _y) ^ ")\n");
		   inner all_sofar (Appeared({ x = _x; y = _y }) :: sofar) id)) true
	 | _ -> failwith "bad input from server")
  in
    if not (nextc chan = '#') then
      failwith "bad input from server";
    inner [] [] (read_int chan) false

let apply_updates s (id,updates) =
  let bot = (try
	       Hashtbl.find s.robots id
	     with
		 Not_found ->
		   let bot = { robot_id = id; robot_pos = None } in
		     Hashtbl.add s.robots id bot;
		     bot) in
    List.iter (fun msg -> update_robot_state s bot msg) updates

let read_packages_here chan =
  let eol = ref false in
  let read_int chan =
    let rec inner sofar =
      let c = input_char chan in
	if c >= '0' && c <= '9' then
	  inner (sofar * 10 + ((Char.code c) - (Char.code '0')))
	else if c = '\n' then
	  (eol := true;
	   sofar)
	else
	  sofar
    in
      inner 0
  in
  let rec inner sofar =
    (try
       if !eol then
	 (print_string "End of packages.\n";
	  List.rev sofar)
       else
	 let pkg_id = read_int chan in
	   if !eol then
	     (print_string "End of packages.\n";
	      List.rev sofar)
	   else
	     let _x = read_int chan in
	     let _y = read_int chan in
	     let weight = read_int chan in
	       print_string ("Package " ^ (string_of_int pkg_id) ^ " is here; dest = (" ^ (string_of_int _x) ^
			     "," ^ (string_of_int _y) ^ ") weight = " ^ (string_of_int weight) ^ "\n");
	       let pkg = { package_id = pkg_id;
			   weight = Some(weight);
			   dest = Some({ x = _x; y = _y });
			   package_pos = None } in
		 inner (pkg :: sofar)
     with
	 Not_found ->
	   failwith "bad input from server\n") in
    print_string "Packages here:\n";
    inner []
;;
print_string "Edsger starting up...\n";;

Random.self_init ();;

let hostname = Sys.argv.(1);;
let port = int_of_string Sys.argv.(2);;

print_string ("Connecting to " ^ hostname ^ ":" ^ (string_of_int port) ^ "\n");;
flush_all ();;
let (s_in,s_out) = Unix.open_connection
		     (Unix.ADDR_INET ((Unix.gethostbyname hostname).Unix.h_addr_list.(0),port));;
print_string "Connected.\n";;

output_string s_out "Player\n";;
flush_all ();;

let (width,height,mapdata) = read_map s_in;;
let map = load_map (width,height,mapdata);;

print_map map;;
print_string "\n\n";;

let my_id = read_int s_in;;
print_string ("my id: " ^ (string_of_int my_id) ^ "\n");;

let capacity = read_int s_in;;
print_string ("my capacity: " ^ (string_of_int capacity) ^ "\n");;

let money = read_int s_in;;
print_string ("my money: " ^ (string_of_int money) ^ "\n\n");;

let homebases (width,height,map) =
  let rslt = ref [] in
    for i = 0 to (height - 1) do
      let row = Array.get map i in
	for j = 0 to (width - 1) do
	  let a_tile = Array.get row j in
	    if a_tile.ty = Home then
	      rslt := a_tile.loc :: !rslt
	done
    done;
    List.rev !rslt

let s =
  {
    width = width;
    height = height;
    map = map;
    my_id = my_id;
    my_pos = { x = max_int; y = max_int };
    my_packages = [];
    capacity = capacity;
    money = money;
    packages = Hashtbl.create 16;
    robots = Hashtbl.create 16;
    my_plan = Attack(0);
    replan = true;
    unvisited_homebases = homebases (width,height,map);
    score = 0
  };;
List.iter (fun x -> apply_updates s x) (read_updates s_in);;

let turncount = ref 0;;

while true do
  incr turncount;
  print_string ("\n********** Turn " ^ (string_of_int !turncount) ^ " **********\n");
  flush_all ();
  let pkgs = read_packages_here s_in in
    update_package_state s pkgs;
    let the_bid = execute_plan s pkgs in
      print_string ("My bid: " ^ (string_of_bid the_bid) ^ "\n");
      make_bid s the_bid s_out;
      flush_all ();
      print_string "Updates: \n";
      List.iter (fun x -> apply_updates s x) (read_updates s_in);
      print_string ("Score: " ^ (string_of_int s.score ^ "\n"))
done
